#ifndef HAVE_CNFG_DEFS
#define HAVE_CNFG_DEFS
typedef struct mac_t {  /* MAC addresses */
        unsigned char device[6];
        unsigned char mcast1[6];
        unsigned char mcast2[6];
        unsigned char mcast3[6];
        unsigned char dflt_srv[6];
} MAC_t;

typedef struct cnfg_t {  /* Configuration elements */
        MAC_t mac;
        unsigned short int ether;
        unsigned short int ext_fifo;
        unsigned short int reset;
        unsigned int vme;
        unsigned short int vme_bto;
        unsigned short int vme_bgto;
} CNFG_t;

typedef CNFG_t *CNFG_ptr;

typedef struct sn_t {  /* Serial Number type structure */
        int status;
        int err_typ;
        int sn;
} SN_t;

enum MAC_ID {DEVICE=0, MCAST1=1, MCAST2=2, MCAST3=3, DFLT_SRV=4, ALL_MACS=8};

enum CR_ID {ETHER=0, EXTFIFO=1, RESET=2, VME=3, BTO=4, BGTO=5, ALL_CRS=8};
enum SET_CLR {CLR=0, SET=1};

#define CNFG_MAXLINE 256

/*
 * Predefine configuration registers
 */
/*
 * MAC Addresses
 */
#define  DEV_MAC      {0x02,0x00,0x00,0x00,0x00,0x00}
#define  BCAST_MAC    {0xFF,0xFF,0xFF,0xFF,0xFF,0xFF}
#define  MCAST1_MAC   {0xFF,0xFF,0xFF,0xFF,0xFF,0xFE}
#define  MCAST2_MAC   {0xFF,0xFF,0xFF,0xFF,0xFF,0xFD}
#define  MCAST3_MAC   {0xFF,0xFF,0xFF,0xFF,0xFF,0xFC}
#define  DFLT_SRV_MAC {0x00,0x0D,0x88,0xB5,0x86,0xAC}
#define  DFLT_FRM_MAC {0xAA,0xBB,0xCC,0xDD,0xEE,0xFF}
/*
 * Ethernet Configuration Register
 */
#define ETH_CR_DFLT   0x0050   //Spontaneous Packets and Protocol Enabled
#define ETH_CR_SPONT  0x0040   //Spontaneous Packet sending enable bit
/*
 * External FIFO Configuration Register
 */
#define FIFO_CR_DFLT    0x0002  //ECC enabled
#define FIFO_CR_TST     0x0001  //Test mode
#define FIFO_CR_ECC     0x0002  //ECC enabled
#define FIFO_CR_INJ     0x0004  //Inject errors
/*
 * Reset Enables/Misc. Configuration Register
 */
#define RST_CR_DFLT     0x0013  //JTAG, Front Panel, Internal
#define RST_CR_INT      0x0001  //Internal
#define RST_CR_FP       0x0002  //Front Panel
#define RST_CR_SRST     0x0004  //VME SYSRESET
#define RST_CR_HR       0x0008  //CMS Hard Reset
#define RST_CR_JTAG     0x0010  //JTAG through PROM CF- pin
#define RST_CR_WARN     0x0020  //Warning packet on shutdown
#define RST_CR_STRTUP   0x0040  //Send packet on startup
/*
 * VME Configuration Register
 */
#define VME_CR_DFLT     0xEDFF1D0F  //Default settings
#define VME_CR_MSTR     0x00000001  //Enable Master
#define VME_CR_SYSCLK   0x00000002  //Enable SYSCLK
#define VME_CR_BTO      0x00000004  //Enable BUS TimeOut
#define VME_CR_ARB      0x00000008  //Enable Arbiter
#define VME_CR_SLV      0x00000010  //Enable Slave (on controller for testing)
#define VME_CR_LOC      0x00000020  //Enable Location Monitor (not impl. yet)
#define VME_CR_IH       0x00000040  //Enable Interrupt Handler
#define VME_CR_USR      0x00000080  //Enable User Defined I/O
#define VME_CR_HIGH     0x00004000  //Enable High order Address and Data
#define VME_CR_FRC_SRST 0x00008000  //Force a VME SYSRESET
#define VME_CR_UPD_IRQ  0x00800000  //Update IRQ Mask (reset to CR value)
#define VME_CR_SRST_OUT 0x40000000  //Enable SYSRESET output to backplane
#define VME_CR_SRST_IN  0x80000000  //Enable SYSRESET input from backplane
#define SET_MSTR_REQ_TYP(t,v)  ((((t)&3)<<8)|((v)&0xFFFFFCFF))
#define SET_MSTR_REQ_LVL(l,v)  ((((l)&3)<<10)|((v)&0xFFFFF3FF))
#define SET_ARB_TYP(t,v)       ((((t)&3)<<12)|((v)&0xFFFFCFFF))
#define SET_IH_LEVELS(l,v)     ((((l)&0x7F)<<16)|((v)&0xFF80FFFF))
#define SET_IH_REQ_TYP(t,v)    ((((t)&3)<<24)|((v)&0xFCFFFFFF))
#define SET_IH_REQ_LVL(l,v)    ((((l)&3)<<26)|((v)&0xF3FFFFFF))
#define SET_IH_DAT_SZ(s,v)     ((((s)&3)<<28)|((v)&0xCFFFFFFF))
/*
 * VME Bus Time Out Register
 */
#define VME_BTO_DFLT  0x30D4  //Default Bus Time Out (200us)
#define VME_BTO_SHRT  0x00D4  //Short Bus Time Out (3.392us)
#define VME_BTO_MAX   0xFFFF  //Maximum Bus Time Out (1.04856ms)
/*
 * VME Bus Grant Time Out Register
 */
#define VME_BGTO_DFLT  0x0C35  //Default Bus Grant Time Out (50.0us)
#define VME_BGTO_MAX   0xFFFF  //Maximum Bus Grant Time Out (1.04856ms)

/*
 * VME Arbiter types
 */
#define ARB_SGL  1 //Single Level Arbiter
#define ARB_PRI  2 //Prioritized Arbiter
#define ARB_RRS  3 //Round Robin Select Arbiter

/*
 * VME Bus Requester types
 */
#define REQ_ROR       0 //Release on Request Requester
#define REQ_RWD       1 //Release When Done  Requester
#define REQ_FAIR_ROR  2 //Fair Release on Request Requester
#define REQ_FAIR_RWD  3 //Fair Release When Done  Requester

extern CNFG_ptr load_cnfg(char *fn);
extern int write_cnfg(CNFG_ptr cp, char *fn);
extern int set_cnfg_dflt(int dflt);
extern int read_cnfg_dflt();
extern int save_cnfg_num(int cnum);
extern int rstr_cnfg_num(int cnum);
extern int read_cnfg_num_dir(int cnum, char **dpp);
extern CNFG_ptr read_cnfg_num_dcd(int cnum);
extern CNFG_ptr read_crs();
extern int set_macs(enum MAC_ID mcid, CNFG_ptr cp);
extern int read_macs_dir(char **dpp);
extern CNFG_ptr read_macs_dcd();
extern int wrt_crs(enum CR_ID crid, CNFG_ptr cp);
extern int set_clr_bits(enum SET_CLR sc, enum CR_ID crid, unsigned int mask);
extern int wrt_ser_num(int ser_num);
extern SN_t rd_ser_num();
extern void cp_macs(CNFG_ptr to, CNFG_ptr from);
extern void cp_crs(CNFG_ptr to, CNFG_ptr from);
extern void cp_cnfg(CNFG_ptr to, CNFG_ptr from);
extern int  cmp_macs(CNFG_ptr to, CNFG_ptr from);
extern int  cmp_crs(CNFG_ptr to, CNFG_ptr from);
extern int  cmp_cnfg(CNFG_ptr to, CNFG_ptr from);
extern void print_macs(CNFG_ptr cp);
extern void print_crs(CNFG_ptr cp);
extern void print_cnfg(CNFG_ptr cp);
extern void log_mac_diffs(CNFG_ptr, CNFG_ptr);
extern void log_cr_diffs(CNFG_ptr, CNFG_ptr);
extern void log_cnfg_diffs(CNFG_ptr, CNFG_ptr);

#endif
