#ifndef HAVE_PKT_INFO
#define HAVE_PKT_INFO
/*
 * Byte offsets into packet
 */
#define MAC_DEST_OFF 0  //Destination MAC Address (6 bytes)
#define MAC_SRC_OFF  6  //Source MAC Address (6 bytes)
#define PKT_LEN_OFF  12 //Packet Length (2 bytes, high order first)
#define HDR_OFF      14 //Header (4 words)
#define PKT_TYP_OFF  15 //Packet type (1 byte)
#define FRAG_NUM_OFF 16 //Packet fragment number (2 words, high order first)
#define PROC_TAG_OFF 16 //Process Tag (5 bits)
#define CMND_ACK_OFF 17 //Command Acknowledged (1 byte)
#define SEQ_PKT_OFF  18 //Sequencial packet ID. (1 word, high order byte first)
#define WRD_CNT_OFF  20 //Word count (13-bits)
#define DATA_OFF     22 //Data offset
#define MSG_OFF      22 //Message offset
#define MAX_DATA_JMB 8994 //Maximum payload for Jumbo packets
#define MAX_DATA_PKT 1494 //Maximum payload for Normal packets




// The following macros assume a base pointer to the packet buffer

/*
 * Header flags and Acknowledge Status in rbuf[HDR_OFF]
 */
#define WITH_DATA(b) ((b[HDR_OFF]>>3)&0x1)
#define AK_STATUS(b) (b[HDR_OFF]&0x7)
#define PKT_FLAGS(b) ((b[HDR_OFF]>>4)&0xF)
#define PROC_TAG(b)  (b[PROC_TAG_OFF]&0x1F)

enum ack_st {NO_ACK =0, CC_S, CC_W, CC_E, CE_I, CIP, CIP_W, CIP_E};
enum pkt_fl {SPONT = 1, FRAG = 2, NEW = 4, PRIO = 8};

union hdr_stat {
  unsigned char full;
  struct {
    unsigned int ackn : 4; //Acknowledge/Status field
    unsigned int spnt : 1; //Spontaneous packet flag
    unsigned int frag : 1; //Packet fragment flag
    unsigned int new  : 1; //New packet flag (1st of a series of fragments)
    unsigned int prio : 1; //Priority packet flag
  } tg; // tag
};

/*
 * Return Packet types in rbuf[PKT_TYP_OFF] (8-bits)
 */
enum pkt_t {NONE_EXP_PKT = -1,
	    NO_DATA_PKT, LPBCK_PKT, TXNWRDS_PKT, EXFIFO_PKT,
            VMED08_PKT, VMED16_PKT, VMED32_PKT, VMED64_PKT,
            JTAG_MOD_PKT, CNFG_MOD_PKT = 10, FLSH_RBK_PKT, ETH_NET_PKT = 16,
            IHD08_STAT_ID_PKT = 0XF8, IHD16_STAT_ID_PKT, IHD32_STAT_ID_PKT,
            INFO_PKT = 0XFD, WARN_PKT, ERR_PKT};

/*
 * Packet Fragment Number is 4 bytes starting at rbuf[FRAG_NUM_OFF] (32-bits)
 */
#define FRAG_NUM(b) (((b[FRAG_NUM_OFF  ]&0xFF)<<24)| \
                     ((b[FRAG_NUM_OFF+1]&0xFF)<<16)| \
                     ((b[FRAG_NUM_OFF+2]&0xFF)<<8) | \
                      (b[FRAG_NUM_OFF+3]&0xFF))

/*
 * Sequencial Packet ID is 2 bytes starting at rbuf[SEQ_PKT_OFF] (16-bits)
 */
#define SEQ_PKT_ID(b) (((b[SEQ_PKT_OFF]&0xff)<<8)|(b[SEQ_PKT_OFF+1]&0xff))

/*
 * Word Count of data returned at rbuf[WRD_CNT_OFF] (13-bits)
 */
#define DAT_WRD_CNT(b) (((b[WRD_CNT_OFF]&0x1f)<<8)|(b[WRD_CNT_OFF+1]&0xff))



/*
 * For Error, Warning, and Info packet types, the source of the
 * information is given by the Source ID in rbuf[MSG_OFF]>>4 (4-bits)
 * The Message Type is given by bits 2 & 3 of rbuf[MSG_OFF]
 * The Message Universal Code Word is bits 0-9 of the word at MSG_OFF.
 */

#define SOURCE_ID(b)    ((b[MSG_OFF]>>4)&0xF)
#define MSG_TYP_ID(b)   ((b[MSG_OFF]>>2)&0x3)
#define UNV_CODE_WRD(b) (((b[MSG_OFF]&0x3)<<8)|(b[MSG_OFF+1]&0xFF))

enum src_id {MISC = 0, VME_CTRL, VME_MSTR, VME_RDBK, VME_IH, VME_SLV,
             VME_ARB, EXFIFO_MOD, ETH_RCV, ETH_TRNS, JTAG_MOD, FLASH_MOD,
             CNFG_MOD, CP_MOD, RST_HNDLR, STRTUP_SHTDWN};
enum typ_id {HDR_INFO = 0, HDR_WARN, HDR_ERROR, HDR_NT_ALLWD};


/*
 * VME Control module Errors/Warnings/Information
 * Source ID = VME_CTRL
 */

// VME Control module Errors
#define EXTRACT_CTRL_TT(b)   (b[MSG_OFF+3]&3)         //2nd Error word Transfer Type
#define EXTRACT_CTRL_DSZ(b)  ((b[MSG_OFF+3]>>2)&3)    //2nd Error word Data Size
#define EXTRACT_CTRL_WRT(b)  ((b[MSG_OFF+3]>>4)&1)    //2nd Error word Write/Read_Bar
#define EXTRACT_CTRL_ASZ(b)  ((b[MSG_OFF+3]>>5)&7)    //2nd Error word Addr. Size
#define EXTRACT_CTRL_DT(b)   (b[MSG_OFF+2]&7)         //2nd Error word Delay Type
#define EXTRACT_CTRL_AT(b)   ((b[MSG_OFF+2]>>3)&0x1F) //2nd Error word Access Types
#define SECND_WRD(b)   (((b[MSG_OFF+2]&0xFF)<<8)|(b[MSG_OFF+3]&0xFF)) //2nd Error word
union ctrl_stat {           // for 2nd Error word.
  unsigned short int full;
  struct {
    unsigned int tt  : 2; //Transfer Type (2 bit field)
    unsigned int dsz : 2; //Data Size     (2 bit field)
    unsigned int wrt : 1; //Write or Read (1 bit field)
    unsigned int asz : 3; //Address Size  (3 bit field)
    unsigned int dt  : 3; //Delay Type    (3 bit field)
    unsigned int at  : 5; //all "access type" bits (5 bit field)
  } tg; // tag
  struct {
    unsigned int     :11; // place holder
    unsigned int pg  : 1; //Program/Data access
    unsigned int sp  : 1; //Supervisory/Non-Privileged access
    unsigned int lk  : 1; // LOCK Command access
    unsigned int cr  : 1; // Configuration ROM/ Control/Status Reg.
    unsigned int ud  : 1; //User Defined Address Modifier
  } fl; // flag
};


/*
 * VME Master Errors/Warnings/Information
 * Source ID = VME_MSTR
 */

// VME Master Errors
#define EXTRACT_MSTR_TT(b)   (b[MSG_OFF+3]&3)         //2nd Error word Transfer Type
#define EXTRACT_MSTR_DSZ(b)  ((b[MSG_OFF+3]>>2)&3)    //2nd Error word Data Size
#define EXTRACT_MSTR_WRT(b)  ((b[MSG_OFF+3]>>4)&1)    //2nd Error word Write/Read_Bar
#define EXTRACT_MSTR_AM(b)  (((b[MSG_OFF+2]&7)<<3)|((b[MSG_OFF+3]>>5)&7)) //2nd Error word Addr. Mod.
#define EXTRACT_MSTR_STATE(b) ((b[MSG_OFF+2]>>3)&0x1F)
union mstr_stat {           // for 2nd Error word.
  unsigned short int full;
  struct {
    unsigned int tt    : 2;
    unsigned int dsz   : 2;
    unsigned int wrt   : 1;
    unsigned int am    : 6;
    unsigned int state : 5;
  } tg; // tag
};
//VME address when error occured is 3rd thru 6th word (64 bit).


/*
 * VME Readback module Errors/Warnings/Information
 * Source ID = VME_RDBK
 */
// VME Readback module Errors
// VME Readback module Warnings
// No second words

/*
 * VME Interrupt Handler module Errors/Warnings/Information
 * Source ID = VME_IH
 */

// VME Interrupt Handler module Errors
#define EXTRACT_IH_IRQ(b)   (b[MSG_OFF+3]&0x7F)     //2nd Error word
#define EXTRACT_IH_ADD(b)   (b[MSG_OFF+2]&7)        //2nd Error word
#define EXTRACT_IH_STATE(b) ((b[MSG_OFF+2]>>4)&0xF) //2nd Error word


// VME Interrupt Handler module Warnings
#define EXTRACT_IRQ_MSK(b)   (b[MSG_OFF+3]&0x7F)



/*
 * External FIFO  Errors/Warnings/Information
 * Source ID = EXFIFO_MOD
 */
// No second words

/*
 * Control Processors Errors/Warnings/Information
 * Source ID = CP_MOD
 */

// Bus Traffic Controller Errors
// No second words


/*
 * Startup/Shutdown Errors/Warnings/Information
 * Source ID = STRTUP_SHTDWN
 */

// Startup/Shutdown Warnings
// No second words

// Startup/Shutdown Information
// No second words


/*
 *Structure for packet response statistics
 */

struct rspn_t {
  int npkt;
  int nspt;
  int nerr;
  int nberr;
  int nbto;
  int nwrn;
  int ninf;
  int nintr;
  int ndat;
  int nack_only;
  int nbad;
  int buf_cnt;
  int buf_typ;
  int totwrds;
  int ackn;
};

#endif
