#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "eth.h"
#include "jtag_defs.h"
#include "codes.h"
#include "pkt_info.h"

extern char *wbuf;
extern int nwbuf;
extern char rbuf[];
extern int nrbuf;


void prg_vcc_prom_ver(char *path, char *ver)
{
  char fname[36];
  char fullname[256];
  int n,pktnum,rslt;
  unsigned int temp_uint;
  static unsigned int device_id = 0xF5057093;
  char *tmp_cp;
  enum pkt_t ptyp;
  enum ack_st ack;

  strcpy(fname,"D783C.V");
  strcat(fname,ver);
  strcat(fname,".mcs");
  strcpy(fullname,path);
  strcat(fullname,"/");
  strcat(fullname,fname);

  rslt = chk_jtag_conn();
  if(rslt == 1){
    temp_uint = read_dev_id();
    if(temp_uint == device_id){
      temp_uint = read_user_code();
      printf("User Code is %08X\n",temp_uint);
      tmp_cp = read_customer_code();
      if(tmp_cp!=0){
	printf("PROM currently programmed with %s\n",tmp_cp);
      }
      else {
	printf("Did not receive customer code. Proceeding anyway\n");
      }
      printf("Erasing PROM...\n");
      rslt = erase_prom();
      if(rslt == 1){
	printf("PROM successfully erased.\n");
	pktnum = 0;
	printf("Reading PROM data file...\n");
	read_mcs(fullname);
	printf("Programming...\n");
	program_prom_cmd();
	send_prg_prom_data();
	send_uc_cc_data(fname);
	while((n=eth_read())>6){
	  pktnum++;
	  ptyp = rbuf[PKT_TYP_OFF]&0xff;
	  if(ptyp>=INFO_PKT){
	    printf("pktnum B%d: %s",pktnum,dcode_msg_pkt(rbuf));
	  } else {
	    printf("pktnum B%d:",pktnum);
	    ack = AK_STATUS(rbuf); 
	    switch(ack){
	    case NO_ACK:
	      printf("Non message or no acknowledge packet received\n");
	      break;
	    case CC_S:
	      printf("Programming completed successfully\n");
	      break;
	    case CC_W:
	      printf("Programming completed with a warning\n");
	      abort_jtag_cmnds();
	      break;
	    case CC_E:
	      printf("Programming completed with an error\n");
	      abort_jtag_cmnds();
	      break;
	    case CE_I:
	      printf("Command execution finished incomplete\n");
	      abort_jtag_cmnds();
	      break;
	    case CIP_W: case CIP_E:
	      printf("Command in progress with an error or warning\n");
	      break;
	    default:
	      printf("Unknown packet returned\n");
	      break;
	    }
	  }
	}
	printf("Verifying...\n");
	verify_prom_cmd();
	send_ver_prom_data();
	printf("Reprogramming FPGA...\n");
	reload_fpga();
      }
      else{
	printf("\a\aPROM is not erased!\n");
      }
    }
    else {
      printf("\a\aWrong device ID -- aborting\n"); 
    }
  }
  else{
    printf("\a\aJTAG connection is bad! Can not proceed.\n");
  }
}
