#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "codes.h"
#include "eth.h"
#include "pkt_info.h"

extern char rbuf[];
extern int nrbuf;

extern FILE *fplog;

#define MAX_MSG_SIZE 1024


char *dcode_msg_pkt(char *buf)
{
  static char *pkt_typ[] = {"INFO","WARN","ERROR","Not a message type"};
  static char *msg_src[] = {"Misc","VME Ctrl","VME Mstr","VME Rdbk",
			    "VME IH","VME Slave","VME Arb","Ext FIFO",
			    "Eth Rcvr","Eth Trns","JTAG mod","Flash mod",
			    "Config mod","CP mod","Rst Hndlr","StrtupShdwn"};
  static char *msg_typ[] = {"INF","WRN","ERR","NAL"};
  static char *unknwn = "Unknown message";
  static char *trns_typ[] = {"SNGL", "BLOCK", "RMW", "UNALG"};
  static char *data_sz[]   = {"D08", "D16", "D32", "D64"};
  static char *addr_sz[]   = {"Axx","A16","A24","A32","A40","A64","Axx","Axx"};
  static char *rdwrt[]     = {"Read","Write"};
  static char *dly_type[]  = {"No_Dly","D4nsX16","D16nsX16","D16usX16",
			      "D4nsX32","D16nsX32","D16usX32","undefined"};
  static char *mstr_st[]   = {"M_idle","Capture1","Req_DTB","R1","No_Exe",
			      "Not_Used","Not_Used","Not_Used","Not_Used",
			      "Wait_4_DTB",
                              "Start","Add_Phase","Wait_4_DTACK_Rel","Sngl_Block",
                              "RMW_1","RMW_Read","RMW_2","RMW_Write","M_Read",
                              "M_Write","Cycle_Done","Next_Cycle","Xfer_Done",
                              "Load_Delay","Delay","Ld_Err_Typ","Bus_Error"};
  static char *ih_st[]     = {"IH_idle","IH_Cap_IRQ","IH_Req_DTB","IH_Wait4DTB",
                              "IH_Start","IACK_Cycle","Status_ID","IACK_Done",
                              "IH_Ld_Err","IH_Bus_Err","IH_Rel_BR","Mask_Intr",
                              "IH_Ld_Warn","IH_Warn"};

  static char *acc_typ_pg[] = {"DATA","PRGM"};
  static char *acc_typ_sp[] = {"NPRV","SUPV"};
  static char *acc_typ_lk[] = {"    ","LOCK"};
  static char *acc_typ_cr[] = {"    ","CCSR"};
  static char *acc_typ_ud[] = {"    ","UDAM"};


  static char msg[MAX_MSG_SIZE];
  static char tmp[MAX_MSG_SIZE];

  union hdr_stat  unh;
  union ctrl_stat unc;
  union mstr_stat unm;

  int n,i;
  enum pkt_t ptyp;
  enum ack_st ack_stat;
  enum pkt_fl pkt_flags;
  enum src_id srcid;
  enum typ_id typid;

  int ih_state,ih_irq,ih_add,curmsk;


  unsigned short int proc_tag, cmnd_ack, seq_id, wrd_cnt;
  unsigned short int code;
  struct ucw *ucptr;

  msg[0]='\0';
  ptyp = buf[PKT_TYP_OFF]&0XFF;
  if(ptyp >= INFO_PKT){

    /* Header info */
    unh.full = buf[HDR_OFF];
    ack_stat = AK_STATUS(buf);
    pkt_flags = PKT_FLAGS(buf);
    proc_tag = PROC_TAG(buf);
    cmnd_ack = buf[CMND_ACK_OFF]&0xFF;
    seq_id   = SEQ_PKT_ID(buf);
    wrd_cnt  = DAT_WRD_CNT(buf);
    /* Message info */
    srcid = SOURCE_ID(buf);
    typid = MSG_TYP_ID(buf);
    code = UNV_CODE_WRD(buf);
    ucptr = (struct ucw *)ptr_bin_srch((int)code, ucwtab, nerrcodes);

    /* Start forming message */
    if(ucptr != NULL){
      sprintf(msg,"%3s: 0x%03X: %-15s: %s\n",ucptr->type,ucptr->code,ucptr->mnem,ucptr->msg);
    /* Check for additional information */
      switch(srcid){
      case VME_CTRL:
	switch(code){
	case VC_MTEr_Fifo:
	case VC_RdEr_Units:
	case VC_RdEr_CtrlWrd:
	  break;
	default:
          strcat(msg,"Additional information:\n");
	  unc.full=SECND_WRD(buf);
	  sprintf(tmp,"\tTransfer Type: %-3s:%-3s:%-5s:%-5s\n",addr_sz[unc.tg.asz],
		  data_sz[unc.tg.dsz],trns_typ[unc.tg.tt],rdwrt[unc.tg.wrt]);
	  strcat(msg,tmp);
	  sprintf(tmp,"\t  Access Type: %4s+%4s+%4s+%4s+%4s\n",acc_typ_sp[unc.fl.sp],
		  acc_typ_pg[unc.fl.pg],acc_typ_lk[unc.fl.lk],
		  acc_typ_cr[unc.fl.cr],acc_typ_ud[unc.fl.ud]);
	  strcat(msg,tmp);
	  sprintf(tmp,"\t   Delay Type: %s\n",dly_type[unc.tg.dt]);
	  strcat(msg,tmp);
	  break;
	}
	break;
      case VME_MSTR:
        strcat(msg,"Additional information:\n");
	unm.full=SECND_WRD(buf);
	sprintf(tmp,"\t    Master State: 0x%02X: %-16s\n",unm.tg.state,mstr_st[unm.tg.state]);
	strcat(msg,tmp);
	sprintf(tmp,"\t   Transfer Type: %-3s:%-5s:%-5s\n",data_sz[unm.tg.dsz],
		trns_typ[unm.tg.tt],rdwrt[unm.tg.wrt]);
	strcat(msg,tmp);
	sprintf(tmp,"\tAddress Modifier: %02X\n",unm.tg.am&0xFF);
	strcat(msg,tmp);
	strcat(msg,"\t  64 bit Address: "); 
	for(i=0;i<8;i++)sprintf(&tmp[2*i],"%02X",buf[MSG_OFF+4+i]&0xFF);
	strcat(msg,tmp);
	strcat(msg,"\n");
	break;
      case VME_IH:
        strcat(msg,"Additional information:\n");
	switch(ptyp){
	case ERR_PKT:
	  ih_state=EXTRACT_IH_STATE(buf);
	  ih_irq=EXTRACT_IH_IRQ(buf);
	  ih_add=EXTRACT_IH_ADD(buf);
	  sprintf(tmp,"\t       IH State: 0x%01X: %-16s\n",ih_state,ih_st[ih_state]);
	  strcat(msg,tmp);
	  sprintf(tmp,"\tIH IRQs pending: ");
	  strcat(msg,tmp);
	  for(i=0;i<7;i++){
	    sprintf(&tmp[i],"%01d",(ih_irq>>(6-i))&1);
	  }
	  strcat(msg,tmp);
	  strcat(msg,"\n");
	  sprintf(tmp,"\tIH IACK Address: %1d\n",ih_add);
	  strcat(msg,tmp);
	  break;
	default:
	  curmsk=EXTRACT_IRQ_MSK(buf);
	  strcat(msg,"\tIH Current IRQ Mask: ");
	  for(i=0;i<7;i++){
	    sprintf(&tmp[i],"%01d",(curmsk>>(6-i))&1);
	  }
	  strcat(msg,tmp);
	  strcat(msg,"\n");
	  break;
	}
	break;
      case ETH_RCV:
        strcat(msg,"Additional information:\n");
	unm.full=SECND_WRD(buf);
	sprintf(tmp,"\t      Command : 0x%02X\n",cmnd_ack&0xFF);
	strcat(msg,tmp);
	sprintf(tmp,"\t  Process Tag : 0x%02X\n",proc_tag&0xFF);
	strcat(msg,tmp);
	sprintf(tmp,"\tSequential ID : 0x%04X\n",seq_id&0xFFFF);
	strcat(msg,tmp);
	sprintf(tmp,"\t   Word Count : 0x%04X (%d)\n",wrd_cnt&0xFFFF,wrd_cnt);
	strcat(msg,tmp);
	break;
      case JTAG_MOD:
        strcat(msg,"Additional information:\n");
	sprintf(tmp,"\t       Status : 0x%04hX\n",SECND_WRD(buf));
	strcat(msg,tmp);
	break;
      default:
/*          strcat(msg,"Additional information:\n"); */
/*  	unm.full=SECND_WRD(buf); */
/*  	sprintf(tmp,"\t      Command : 0x%02X\n",cmnd_ack&0xFF); */
/*  	strcat(msg,tmp); */
/*  	sprintf(tmp,"\t  Process Tag : 0x%02X\n",proc_tag&0xFF); */
/*  	strcat(msg,tmp); */
/*  	sprintf(tmp,"\tSequential ID : 0x%04X\n",seq_id&0xFFFF); */
/*  	strcat(msg,tmp); */
/*  	sprintf(tmp,"\t   Word Count : 0x%04X (%d)\n",wrd_cnt&0xFFFF,wrd_cnt); */
/*  	strcat(msg,tmp); */
/*          sprintf(tmp,"\t       Source : %01X\n",srcid&0xF); */
/*          strcat(msg,tmp); */
/*          sprintf(tmp,"\t    Msg. Type : %01X\n",typid&0xF); */
/*          strcat(msg,tmp); */
/*          sprintf(tmp,"\t         Code : %03X\n",code&0xFFF); */
/*          strcat(msg,tmp); */
/*          for(i=0;i<DATA_OFF+2*wrd_cnt;i++){ */
/*            printf("%02X",buf[i]&0xFF); */
/*          } */
/*          printf("\n"); */
	break;
      }
    }
    else {
      sprintf(msg,"%3s: 0x%03X: %-15s: Unmatched code 0x%03X is undefined.\n",msg_typ[typid],code,msg_src[srcid],code);
      strcat(msg,"Additional information:\n");
      if((pkt_flags & SPONT)!=0){
        strcat(msg,"\tThis is a spontaneous packet\n");
      }
      else {
        sprintf(tmp,"\t     Status: %-6s: %s\n",ak_status[ack_stat].mnem,ak_status[ack_stat].status);
        strcat(msg,tmp);
        sprintf(tmp,"\tProcess Tag: 0x%02X\n",proc_tag);
        strcat(msg,tmp);
        sprintf(tmp,"\tCommand Ack: 0x%02X\n",cmnd_ack);
        strcat(msg,tmp);
        sprintf(tmp,"\tSeq.Pkt.ID : %d\n",seq_id);
        strcat(msg,tmp);
      }
    }
  }
  else {
    strcat(msg,"Not a message packet type");
  }
  strcat(msg,"\n");
  return msg;
}

void *ptr_bin_srch(int code, struct ucw *arr, int n)
{
  struct ucw *low  = &arr[0];
  struct ucw *high = &arr[n-1];
  struct ucw *mid;

  while(low < high){
    mid = low + (high-low)/2;
    if(code < mid->code){
      high = mid;
    }
    else if(code > mid->code){
      low = mid;
    }
    else {
      return mid;
    }
  }
  return NULL;
}

struct rspn_t flush_pkts()
{
  int i,n;
  union hdr_stat hdr;
  enum pkt_t pkt_type;
  enum src_id srcid;
  enum ack_st ackn;

  int spont,frag,new,prio,data;
  int maxbytes,maxwrds,maxints,maxlongs;
  int numwrds;
  unsigned char *cp;
  unsigned short int *hp;
  unsigned short int code;
  unsigned int *ip,*lph,*lpl;

  unsigned char opcode;
  struct ucw *fopptr;
  struct ucw *pktptr;
  struct rspn_t rsp;

  rsp.npkt=0;
  rsp.nspt=0;
  rsp.nerr=0;
  rsp.nberr=0;
  rsp.nbto=0;
  rsp.nwrn=0;
  rsp.ninf=0;
  rsp.nintr=0;
  rsp.ndat=0;
  rsp.nack_only=0;
  rsp.nbad=0;
  rsp.buf_cnt=0;
  rsp.buf_typ=0;
  rsp.totwrds=0;
  rsp.ackn=0;

  while((n=eth_read())>6){
    rsp.npkt++;
    hdr.full = rbuf[HDR_OFF];
    spont = hdr.tg.spnt;
    frag  = hdr.tg.frag;
    new   = hdr.tg.new;
    prio  = hdr.tg.prio;
    ackn  = AK_STATUS(rbuf);
    data  = WITH_DATA(rbuf);
    pkt_type = rbuf[PKT_TYP_OFF] & 0xFF;
    opcode = rbuf[CMND_ACK_OFF];
    fopptr = (struct ucw *)ptr_bin_srch((int)opcode,foptab,nopcodes);
    pktptr = (struct ucw *)ptr_bin_srch((int)pkt_type,pkttab,npktcodes);
    numwrds = DAT_WRD_CNT(rbuf);
    if(spont){
      rsp.nspt++;
      if(pkt_type>=INFO_PKT){
        printf("    Message from controller:\n");
        printf("    %s", dcode_msg_pkt(rbuf));
        if(fplog != NULL){
          fprintf(fplog,"    Message from controller:\n");
          fprintf(fplog,"    %s", dcode_msg_pkt(rbuf));
	}
        code = UNV_CODE_WRD(rbuf);
        switch(pkt_type){
	case INFO_PKT:
	  rsp.ninf++;
	  break;
	case WARN_PKT:
	  rsp.nwrn++;
	  break;
	case ERR_PKT:
 	  rsp.nerr++;
	  if(code == VM_BERR_Slv || code == VI_BERR_Slv){
            rsp.nberr++;
          }
          if(code == VM_BTO || code == VI_BTO){
            rsp.nbto++;
	  }
	  break;
	default:
	  break;
	}
      }
      else if(pkt_type>=IHD08_STAT_ID_PKT && pkt_type<=IHD32_STAT_ID_PKT){
        rsp.nintr++;
        rsp.totwrds += numwrds;
      }
      else {
	printf("Unexpected Spontaneous packet type: %02X\n",pkt_type&0xFF);
      }
    }
    else if(ackn >0 && data ==0){
      rsp.nack_only++;
      rsp.ackn=ackn;
      printf("Acknowledge packet:\n  Status:%-6s:%20s %s\n",ak_status[ackn].mnem,fopptr->mnem,ak_status[ackn].status);
    }
    else if(numwrds > 0){
      rsp.ndat++;
      rsp.totwrds += numwrds;
      printf("Flushing %d of data from a %s packet\n",numwrds,pktptr->mnem);
    }
    else {
      rsp.nbad++;
      printf("Illegal Header encountered\n");
    }
  }
  return rsp;
}
