#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include "codes.h"
#include "slave_dcl.h"
#include "slave_dev.h"
#include "cnfg_defs.h"
#include "test_defs.h"
#include "eth.h"
#include "eth_ext_dcl.h"


int pre_prod;
int ser_num,mac_sn;
SN_t sn;
int bcast,uninit,slave,master,tmp_master;
FILE *fplog;

#define   SLV_ETH_CR_DFLT     0x0050
#define   SLV_FIFO_CR_DFLT    0x0002
#define   SLV_RST_CR_DFLT     0x0013
#define   SLV_VME_CR_DFLT 0x2E000190
#define   SLV_VME_BTO_DFLT    0x0000
#define   SLV_VME_BGTO_DFLT   0x0000

CNFG_t std_cnfg = { {DEV_MAC,
                     MCAST1_MAC,
                     MCAST2_MAC,
                     MCAST3_MAC,
                     DFLT_SRV_MAC},
                    ETH_CR_DFLT,
                    FIFO_CR_DFLT,
                    RST_CR_DFLT,
                    VME_CR_DFLT,
                    VME_BTO_DFLT,
                    VME_BGTO_DFLT};

CNFG_t std_slv_cnfg = { {DEV_MAC,
			 MCAST1_MAC,
			 MCAST2_MAC,
			 MCAST3_MAC,
			 DFLT_SRV_MAC},
			SLV_ETH_CR_DFLT,
			SLV_FIFO_CR_DFLT,
			SLV_RST_CR_DFLT,
			SLV_VME_CR_DFLT,
			SLV_VME_BTO_DFLT,
			SLV_VME_BGTO_DFLT};

int keep_running;

void int_hndlr(int sig_num)
{
  switch(sig_num){
    case SIGINT:
      keep_running = 0;
      break;
    default:
      printf("    Unknown signal!\n");
      break;
  }
}

int main(int argc,char *argv[])
{
  static char *undrscr = "----------------------------\n";
 
  unsigned char dflt_mac[] = DFLT_FRM_MAC;
  unsigned char bcast_mac[] = BCAST_MAC;
  unsigned char mstr_mac[] = DEV_MAC;
  char ans[128];
  char *cp;
  int i,j,l,n,tst_tot,fail_tot,lcnt,syserr;
  int rslt,tst_err,tst_err_cnt,diffs;
  int slv_mtch,unt_mtch,rsp_macs,unt,rsp,mac_tried,tmp_mac;
  int npol;
  unsigned char slv_mac[6];
  unsigned char pmacs[NMAX_MACS][6];
  int tag_macs[NMAX_MACS];
  FILE *fp;
  time_t tp;
  char logfn[128];
  char *tmp;
  CNFG_t *rcv_cnfg;

  /*
   *
   * Initialization
   */

  pre_prod=0;
  keep_running=1;
  signal(SIGINT,int_hndlr);

  fp=fopen("mac_addr_slave.hex","r");
  fscanf(fp,"%02X-%02X-%02X-%02X-%02X-%02X",&slv_mac[0],&slv_mac[1],&slv_mac[2],&slv_mac[3],&slv_mac[4],&slv_mac[5]);
  fclose(fp);

  for(i=0;i<JMB_DATA_LEN;i++)wbuf[i]=0;
  eth_open("vme0");
  bcast  = eth_register_mac(bcast_mac);
  uninit = eth_register_mac(dflt_mac);
  slave  = eth_register_mac(slv_mac);
  eth_reset();
  eth_read_timeout(RD_TMO_short);
  cur_mac = bcast;
  npol = poll_macs(pmacs);
  slv_mtch=unt_mtch=rsp_macs=0;
  /*
   *
   * Scanning for boards
   */
  printf("received %d packets\n",npol);
  if(npol>0){
    printf("The following MAC devices were found:\n");
    for(n=0;n<npol;n++){
          printf("%d: ",n);
          print_mac(pmacs[n]);
      if(maccmp(pmacs[n],slv_mac)==0){
        slv_mtch++;
        tag_macs[n]=slave;
      }
      else if(maccmp(pmacs[n],dflt_mac)==0){
        unt_mtch++;
        unt=n;
        tag_macs[n]=uninit;
      }
      else {
        tag_macs[n]=0;
        rsp=n;
        rsp_macs++;
      }
    }
    if(slv_mtch>0){
      printf("Slave found:\n");
      for(n=0;n<npol;n++){
        if(tag_macs[n]==slave){
          printf("%d: ",n);
          print_mac(pmacs[n]);
        }
      }
      printf("Checking slave for proper configuration\n");
      tmp_mac=cur_mac;
      cur_mac=slave;{
	rcv_cnfg = read_cnfg_num_dcd(0);
	diffs = cmp_crs(&std_slv_cnfg,rcv_cnfg);
	if(diffs>0){
	  printf("This device is not configured as a slave!\n");
	  printf("Reconfiguring with slave defaults.\n");
	  wrt_crs(ALL_CRS,&std_slv_cnfg);
	  save_cnfg_num(0);
	  set_cnfg_dflt(0);
	}
        free(rcv_cnfg);
      }
      cur_mac=tmp_mac;
    }
    if(rsp_macs+unt_mtch>1){
      printf("The following MAC devices are possible masters:\n");
      for(n=0;n<npol;n++){
        if(tag_macs[n]==0 || tag_macs[n]==uninit){
          printf("%d: ",n);
          print_mac(pmacs[n]);
        }
      }
      printf("There should only be one master!\n");
      printf("Check your setup and try again.\n");
      eth_close();
      exit(1);
    }
    else if(unt_mtch>0){
      cur_mac = uninit;
      mac_tried =unt;
    }
    else if(rsp_macs>0){
      tmp_master = eth_register_mac(pmacs[rsp]);
      cur_mac = tmp_master;
      mac_tried = rsp;
    }
    else{
      printf("No Master found!\n");
      printf("Check your setup and try again.\n");
      eth_close();
      exit(1);
    }
  }
  /*
   *
   * Check serial number
   */
  printf("Checking the serial number.\n");
  sn = rd_ser_num();
  if(sn.status==-1){
    printf("The board did not respond. MAC: ");
    print_mac(pmacs[mac_tried]);
    eth_close();
    exit(1);
  }
  else if(sn.status==-2){
    printf("Received error type %2X from controler\n",sn.err_typ);
    eth_close();
    exit(1);
  }
  else {
    ser_num = sn.sn;
  }
  if(ser_num==-1){
    printf("This board is either uninitialized or pre-production #1.\n");
    printf("Enter Serial Number : ");
    scanf("%d",&ser_num);
  }
  else {
    mac_sn = (int) ((signed char)pmacs[mac_tried][3]);// sign extension
    mac_sn = (mac_sn&0xFF000000)|((pmacs[mac_tried][3]&0xFF)<<16)|((pmacs[mac_tried][4]&0xFF)<<8)|(pmacs[mac_tried][5]&0xFF);
    if(ser_num != mac_sn){
      printf("The serial number (%d) is not consistent with the MAC address (%d).\n",ser_num,mac_sn);
      printf("MAC: ");
      print_mac(pmacs[mac_tried]);
      printf("Is this the correct serial number? (y/n): ");
      scanf("%s",ans);
      if(ans[0]=='n'){
        printf("Enter Serial Number (neg. for pre-prod.): ");
        scanf("%d",&ser_num);
      }
      else{
        printf("OK, continuing to use Serial Number %d:\n",ser_num);
      }
    }
    else {
      printf("This board has been previously initialized.\n");
      printf("The MAC address is consistent with the serial number.\n");
      printf("MAC: ");
      print_mac(pmacs[mac_tried]);
      printf("Serial Number: %d\n",ser_num);
    }
  }
  if(ser_num < 0)pre_prod=1;
  mstr_mac[3]=(ser_num>>16)&0xFF;
  mstr_mac[4]=(ser_num>>8)&0xFF;
  mstr_mac[5]=ser_num&0xFF;
  master = eth_register_mac(mstr_mac);
  printf("The MAC address will be set to : ");
  print_mac(mstr_mac);
  for(i=0;i<6;i++)std_cnfg.mac.device[i]=mstr_mac[i];

  /*
   * Start logging
   */
  if(ser_num<0){
    sprintf(logfn,"logs/VCC_pp%02d_tests.log",-ser_num);
  }
  else{
    sprintf(logfn,"logs/VCC_%02d_tests.log",ser_num);
  }
  if((fplog = fopen(logfn,"a"))==0){
    printf("Could not open log file. %s\n",logfn);
    if(errno==2){
      printf("No logs directory found\n");
      printf("Attempting to create it\n");
      syserr = system("mkdir logs");
      if((fplog = fopen(logfn,"a"))==0){
        printf("Still could not open log file.\nSystem error %d\nError number %d\n",syserr,errno);
        perror("dbg_test: ");
        exit(1);
      }
      else {
        printf("Success: log file is open\n");
      }
    }
    else {
      printf("Could not open log file.\nError number %d\n",errno);
      perror("dbg_test: ");
      exit(1);
    }
  }
  fprintf(fplog,"\n\n\n=============== New Entry (ver 4.0) =========\n",ser_num);
  rslt = time(&tp);
  if(rslt==-1){
    fprintf(fplog,"\nTime was unavailable.\n");
  }
  else {
    fprintf(fplog,"\n%s\n",ctime(&tp));
  }
  fflush(fplog);

  /*
   *
   * Start of testing
   */
  printf("\n*****************************************\n");
  printf("\nStart of debugging: Board Number %02d\n",ser_num);
  fprintf(fplog,"*****************************************\n\n");
  fprintf(fplog,"Start of debugging: VCC Board #%02d\n\n",ser_num);
  fprintf(fplog,"*****************************************\n");
  fflush(fplog);
  printf("\n\n");
  fprintf(fplog,"\n\n");
  fprintf(fplog,undrscr);

  // Initialization for VME testing

  // set configuration to std_cnfg
  //Write standard configuration to all CR's
  printf("    Writing CRs:\n");
  wrt_crs(ALL_CRS,&std_cnfg);

  printf("  Testing VME:\n");
  set_VME_mode();
  prst_ff();
  init_slave();

  set_clr_bits(SET, VME, VME_CR_HIGH); // Enable high order bits.

  tst_tot=0;
  tst_err_cnt=0;
  fail_tot=0;
  while(keep_running){
    //    tst_err = vme_berr();
    //    scanf("%s",ans);
    tst_err = vme_dly();
    //    usleep(10000);
    tst_err_cnt += tst_err;
    if(tst_err)fail_tot++;
    tst_tot++;
    printf("\nFailure rate %d/%d = %5.1f\%\n",fail_tot,tst_tot,100*(double)fail_tot/tst_tot);
  }
  if(fail_tot){
    printf("********\n\nFailed %d out of %d tests.\n",fail_tot,tst_tot);
    printf("With a total of %d errors detected.\n\n********\n",tst_err_cnt);
  }
  else{
    printf("********\n\nPassed all %d tests.\n\n********\n",tst_tot);
  }
  eth_close();
  fclose(fplog);
}

