#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h> 
#include <netinet/if_ether.h> 
#include <netinet/in.h> 
#include <netinet/ip.h> 
#include <net/if.h> 
#include <sys/ioctl.h> 
#include <netpacket/packet.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include "eth.h"

int  eth_open(char *dev_name);
int  eth_register_mac(unsigned char *mac);
void eth_close(void);
int  eth_reset(void);
int  eth_read();
int  eth_write();
int  eth_enableblock(void);
int  eth_disableblock(void);

extern unsigned char ebuf[];
extern unsigned char *wbuf;
extern unsigned char rbuf[];
extern int nwbuf;
extern int nrbuf;
extern struct ethhdr *ether_header[];
extern unsigned char hw_source_addr[]; 
extern int nmacs;
extern int cur_mac;

static int fd_schar;

#ifdef SIM_OUTPUT
static  FILE *smfp;
#endif
static char *fmt;
static char dsh[] = "%02X-"; 
static char nodsh[] = "%02X\n"; 

int eth_open(char *dev_name)
{
  int i;
  char *dev;
  if((dev=(char *)malloc((6+strlen(dev_name))*sizeof(char)))==0){
    printf("Device name allocation error!\n");
    return -1;
  }
  sprintf(dev,"/dev/%s",dev_name);
  fd_schar = open(dev, O_RDWR);
  if (fd_schar == -1) {
    perror("open");
    return -1;
  }
  free(dev);
  if(ioctl(fd_schar,SCHAR_GET_MAC,hw_source_addr)){
    printf("Error reading MAC address\n");
  }
  printf("NIC MAC addr.  (source): ");
  for(i=0;i<6;i++){
    fmt = (i<5)?dsh:nodsh;
    printf(fmt,hw_source_addr[i]&0xff);
  }
  nmacs=0;
#ifdef SIM_OUTPUT
  smfp=fopen("sim_out.hex","w");
#endif
}

int eth_register_mac(unsigned char *mac)
{
  int  i,mac_idx;
  mac_idx = nmacs++; 
  for(i=1;i<6;i++){
   hw_source_addr[i]=mac_idx;
  }
  if(nmacs>NMAX_MACS){
    printf("Exceeded maximum allowed MAC destinations\n");
    return -1;
  }
  if((ether_header[mac_idx]=(struct ethhdr *)malloc(sizeof(struct ethhdr)))==0){
    printf("Could not register MAC address: Allocation error\n");
    return -1;
  }
  memcpy(ether_header[mac_idx]->h_dest, mac, ETH_ALEN);
  memcpy(ether_header[mac_idx]->h_source,hw_source_addr, ETH_ALEN);
  printf("Registered Dest. MAC %2d: ",mac_idx);
  for(i=0;i<6;i++){
    fmt = (i<5)?dsh:nodsh;
    printf(fmt,ether_header[mac_idx]->h_dest[i]&0xff);
  }
  return mac_idx;
}

int eth_reset(void)
{ 
  if(ioctl(fd_schar,SCHAR_RESET)==-1){
    printf(" error in SCHAR_RESET \n");
  }

  return 0;
}

int eth_read(int count,char *buf)
{
int size;
int lp;
  size=read(fd_schar,rbuf,nrbuf);
  for(lp=0;size>=0 && size <7 && rbuf[0]==0x03 && lp<100;lp++){
    usleep(100);
    size=read(fd_schar,rbuf,nrbuf);
  }
  return size;
}


int eth_write()
{
  int pkt_size,pkt_stp;
  int i;
  //  check for registered MAC address
  if(cur_mac<0 || cur_mac >= nmacs){
    printf("Your using an unregister MAC address\n");
    return -99;
  }
  //Creating the packet
  ether_header[cur_mac]->h_proto = htons(nwbuf);/* Length of data */
  memcpy(ebuf, ether_header[cur_mac], ETH_HLEN);
  pkt_size = ETH_HLEN + nwbuf;
  //  memcpy(ebuf+ETH_HLEN, wbuf, nwbuf); /*don't need this extra copy now*/
#ifdef SIM_OUTPUT
#ifdef SIM_TRUNK_MAX
  pkt_stp = SIM_TRUNK_MAX;
#else
  pkt_stp = pkt_size;
#endif
  for(i=0;i<pkt_size&&i<pkt_stp;i++)fprintf(smfp,"%02X",ebuf[i]&0xff);
#ifdef SIM_TRUNK_MAX
  if(i== SIM_TRUNK_MAX)  fprintf(smfp,"...");
#endif
  fprintf(smfp,"\n");
  fflush(smfp);
#endif
  return write(fd_schar, (const void *)ebuf, pkt_size);
}


void eth_close(void)
{
  int i;
#ifdef SIM_OUTPUT
  fclose(smfp);
#endif
  for(i=0;i<nmacs;i++){
    free(ether_header[i]);
  }
  close(fd_schar);
}

int eth_read_timeout(int rd_tmo)
{
  if(ioctl(fd_schar,SCHAR_READ_TIMEOUT,rd_tmo)){
    printf("Error in SCHAR_READ_TIMEOUT \n");
  }
  return 0;
}
