#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "pkt_info.h"
#include "codes.h"
#include "test_defs.h"

extern char *wbuf;
extern int nwbuf;
extern char rbuf[];
extern int nrbuf;
short flash_rbk[MAX_PAGES][MAX_ADDRS];

void flash_init_rbk()
{
  int i,j;
  for(i=0;i<MAX_PAGES;i++){
    for(j=0;j<MAX_ADDRS;j++){
      flash_rbk[i][j]=-999;
    }
  }
}


int flash_write(enum CMP_MODE cm)
{

  int i,j,n,nr,nwrds;
  int l,lcnt,cur,nwrite,strt,loop,err_type,type;
  int page,addr,num_pages,icnt,bcnt,bcindex,data;
  int sect;
  unsigned char tst_dat;
  unsigned char wrt_val[] = {0,0,0xFF,0xAA,0x55,0,0xFF};

  if(cm == NO_CMP){
    return -1; //cmp mode "NO_CMP" is "no comparison" (read only)
  }
  // initialize buffers
  for(i=0;i<9000;i++)wbuf[i]=i&0xff;
  for(i=0;i<9000;i++)rbuf[i]=i&0xff;
  for(sect=BEG_SECT;(sect<=END_SECT)&&(sect<MAX_SECT);sect++){
    num_pages = 0;
    nwrds=0;
    icnt = 4;
    // Write to Flash
    for(page=8*sect;page<8*sect+NUM_PGS;page++){
      wbuf[icnt++]=WRT | (page&0x7F);  //page to write plus read/write flag
      bcindex=icnt;
      icnt++;
      bcnt=0;
      for(addr=BEG_ADDRS;(addr<=END_ADDRS)&&(addr<MAX_ADDRS);addr++){
        switch(cm){
	case INC_CMP:
          tst_dat = (((page&3)<<6)|(addr&0x3F));
          break;
        case DEC_CMP:
          tst_dat = (0xFF-(((page&3)<<6)|(addr&0x3F)));
          break;
        default:
          tst_dat  = wrt_val[cm];
          break;
        }
        wbuf[icnt++]=addr&0x3F;      //address in page
        wbuf[icnt++]=tst_dat; //data to write to address
        bcnt++;
        nwrds++;
      }
      wbuf[bcindex]=bcnt;
      nwrds++;
      num_pages++;
    }
    wbuf[2]=0x00;
    wbuf[3]=num_pages;
    nwrds++;
    flash_cmds(nwrds);
    printf("    Writing Section %d:\n",sect);
    usleep(10000*NUM_PGS);//wait 10mS/page
  }
  return 0;
}

int flash_read(enum CMP_MODE cm)
{
  int i,j,n,nr,nwrds,no_data,data_chng;
  int l,lcnt,cur,nwrite,strt,loop,err_type,type,errcnt;
  int page,addr,num_pages,icnt,bcnt,bcindex,data,sect;
  int toterrs;
  // initialize buffers
  for(i=0;i<9000;i++)wbuf[i]=i&0xff;
  for(i=0;i<9000;i++)rbuf[i]=i&0xff;

  toterrs=0;
  no_data=0;
  for(sect=BEG_SECT;(sect<=END_SECT)&&(sect<MAX_SECT);sect++){
    num_pages = 0;
    nwrds=0;
    icnt = 4;
    // Read from Flash
    for(page=8*sect;page<8*sect+NUM_PGS;page++){
      wbuf[icnt++]=RD | (page&0x7F);  //page to read plus read/write flag
      bcindex=icnt;
      icnt++;
      bcnt=0;
      for(addr=BEG_ADDRS;(addr<=END_ADDRS)&&(addr<MAX_ADDRS);addr++){
        wbuf[icnt++]=addr&0x3F;      //address in page
        wbuf[icnt++]=0x00;      //space holder for read
        bcnt++;
        nwrds++;
      }
      wbuf[bcindex]=bcnt;
      nwrds++;
      num_pages++;
    }
    wbuf[2]=0x00;
    wbuf[3]=num_pages;
    nwrds++;
    flash_cmds(nwrds);
    nr=eth_read();
    //  printf("read %d words\n",n);
    if(nr>6){
      data_chng = no_data;
      while((errcnt = dcode_flash_pkt(cm))<0 && (data_chng == no_data)){
        nr=eth_read();
        if(nr<=6){
	  printf("    No additional packets from controller! Section %2d\n",sect);
          no_data--;
        }
      }
    }
    else {
      printf("    No packet from controller! Section %2d\n",sect);
      return -1;
    }
    if(errcnt!=0){
      printf("    %3d errors found in section %2d\n",errcnt,sect);
    }
    else if(cm == NO_CMP){
      printf("    Section %2d: Read in\n",sect);
    }
    else{
      printf("    Section %2d: no errors\n",sect);
    }
    toterrs+=errcnt;
  }
  return (no_data <0) ? no_data : toterrs;
}


int dcode_flash_pkt(enum CMP_MODE cm)
{
  int d,i,j,k,l;
  int err_type,type,nwrds,errcnt;
  unsigned char addr,page,bcnt,data,exp,opcode;
  unsigned char exp_val[] = {0,0,0xFF,0xAA,0x55,0,0xFF};
  enum ack_st ackn;
  struct ucw *fopptr;

  //Return values are:
  //  positive (0-n): error count
  //  -1            : Message packet
  //  -2            : Acknowledge only packet
  //  -3            : Wrong packet type
  //  -4            : No status, no data
  d=DATA_OFF;
  type = rbuf[PKT_TYP_OFF]&0xFF;
  nwrds = DAT_WRD_CNT(rbuf);
  ackn = AK_STATUS(rbuf);
  opcode = rbuf[CMND_ACK_OFF];
  fopptr = (struct ucw *)ptr_bin_srch((int)opcode,foptab,nopcodes);
  if(type >= INFO_PKT){
    printf("Message from controller:\n%s\n",dcode_msg_pkt(rbuf));
    return -1;
  }
  else if(ackn>0 && nwrds == 0){
    printf("Acknowledge packet:\n  Status:%-6s:%20s %s\n",ak_status[ackn].mnem,fopptr->mnem,ak_status[ackn].status);
    return -2;
  }
  else if(type != FLSH_RBK_PKT){
    printf("Non Flash packet:\n  Packet type: %02X\n",type);
    return -3;
  }
  else if(ackn == 0 && nwrds == 0){
    printf("No status and no data!\n");
    return -4;
  }
  else {
    err_type = 0;
    errcnt =0;
    i=0;
    //    printf(" Page Addr Data\n");
    while(i<nwrds){
      page = rbuf[d++]&0x7F; 
      bcnt = rbuf[d++];
      i++;
      for(j=0;j<bcnt;j++){
	addr = rbuf[d++]&0x3F;
        data = rbuf[d++];
        switch(cm){
	case INC_CMP:
          exp = (((page&3)<<6)|(addr&0x3F));
          break;
        case DEC_CMP:
          exp = (0xFF-(((page&3)<<6)|(addr&0x3F)));
          break;
        default:
          exp  = exp_val[cm];
          break;
        }
        i++;
        if(cm!=NO_CMP && data!=exp){
	  errcnt++;
	  //          printf("*** error ***  %02X   %02X   rbk: %02X   exp: %02X\n",page&0x7F,addr&0x3F,data&0xFF,exp&0xFF);
        }
        flash_rbk[page][addr]=data&0xFF;
      }
    }
  }
  return errcnt;
}

void flash_print(FILE *fp)
{
  int sc,pg,cl,rw,lprnt;
  for(sc=0;sc<MAX_SECT;sc++){
    fprintf(fp,"Section %02d:\n",sc);
    for(pg=8*sc;(pg<8*sc+NUM_PGS) && (pg<MAX_PAGES);pg++){
      for(rw=0;rw<4;rw++){
        lprnt=0;
        for(cl=0;cl<16;cl++){
          if(flash_rbk[pg][16*rw+cl]!=-999){
            lprnt=1;
          }
        }
        if(lprnt==1){
          fprintf(fp,"%02X:%02X:",pg&0xFF,16*rw);
          for(cl=0;cl<16;cl++){
            if(flash_rbk[pg][16*rw+cl]!=-999){
              fprintf(fp," %02X",flash_rbk[pg][16*rw+cl]);
            }
            else {
              fprintf(fp,"   ");
            }
          }
          fprintf(fp,"\n");
        }
      }
    }
  }
}
