#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "codes.h"
#include "cnfg_defs.h"
#include "slave_dev.h"
#include "slave_dcl.h"
#include "eth.h"
#include "eth_ext_dcl.h"
#include "mac_utils_dcls.h"

char *settings[]  = {"tt","dsz","Dxx","wrt(W)","rd(R)","asz","Axx","dt","at"};
char *s_desc[]  = {":Trns Type",":data size",":D08,D16...",":write",":read",
                   ":addr size",":A16,A24...",":delay type",":access type"};
char *values[] = {"s","dev","off","data","dly","uam","lp"};
char *v_desc[]  = {":Slot",":device",":offset",":data entry",
                   ":delay",":user AM",":Loop count"};
char *cmds[] = {"h","inc(I)","dec(D)","ron","roff","clr","ex","q"};
char *c_desc[]  = {":help",":Inc. Addr",":Dec. Addr",":Ena Reads",":Dis Reads",":Clr Data cnt",":Execute",":quit"};
char *blank = "    ";

int numset = 9;
int numval = 7;
int numcmd = 8;
int max;
unsigned int slot = 11;
int b_addr[] = {0x0000,0x0800,0x1000,0x2000,0x3000,0x3008};

int dev_mask[]  = {0xFFFFFFFF,  //NONE
                   0xFFFFF800,  //A16
                   0xFFFFC000,  //A24
                   0xFFFFC000,  //A32
                   0xFFFFC000,  //A40
                   0xFFFFC000}; //A64

int slot_mask_l[] = {0x00000000,  //NONE
                     0x000007FF,  //A16
                     0x0007FFFF,  //A24
                     0x07FFFFFF,  //A32
                     0xFFFFFFFF,  //A40
                     0xFFFFFFFF}; //A64

int slot_mask_h[] = {0x00000000,  //NONE
                     0x00000000,  //A16
                     0x00000000,  //A24
                     0x00000000,  //A32
                     0x00000007,  //A40
                     0x07FFFFFF}; //A64

int slot_shft[] = { 0,  //NONE
                   11,  //A16
                   19,  //A24
                   27,  //A32
                   35,  //A40
                   59}; //A64

char *dfmt[] = {"%02X","%04hX","%08X","%08X"};
int   dlmt[] = {17,10,5,3};
int cur_am;
unsigned int idata[1024]; 
unsigned char *rp;
int numwrds,dcnt;
int no_rsp,verrfl,numerr_wrds;
int verr_type,isrc;
unsigned char verr_wrds[48];
unsigned char data_wrds[9000];
unsigned int off,aoff;
char *msg_src[]  = {"Misc","VME Ctrl","VME Mstr","VME Rdbk",
                       "VME IH","VME Slave","VME Arb","Ext FIFO",
                       "Eth Rcvr","Eth Trns","JTAG mod","Flash mod",
                       "Config mod","BTC mod","Rst Hndlr","StrtupShdwn"};
char *verr_types[]  = {"No Ctrlerr","Unkn Addr","Unkn Dly","Incomp Opt",
                       "uknmstrerr","BERR","not allwd","bustimeout",
                       "V Rdbk Err","V Rdbk Err","V Rdbk Err","V Rdbk Err",
                       "V IH Err","V IH Err","V IH Err","V IH Err",
                       "unknownerr"};

enum SLV_DEV dev;
enum TRNS_TYP tt;
enum DATA_SZ  dsz; 
enum WRT_RD wrt;
enum ADDR_SZ asz;
enum DLY_TYP dt;
enum etypes {FF_NO_ERR = 0, FF_RD_ERR = 1, FF_MT_ERR = 2, FF_RT_ERR = 4, FF_MK_ERR = 8, FF_WRT_ERR = 16, TC_NO_DATA = 32, TC_NOT_EXC = 64, TC_UNK_FNC = 96, TC_START_UP = 192, TC_RLD_PNDG = 224} err_type;


int bcast,uninit,slave,master,tmp_master;

VCMD_t vc;

FILE *fplog;

void display(int help);
void disp_v();
void disp_h();
int non_hex_digits(char *);
void update_addr();
void update_am();

int main()
{

  unsigned char dflt_mac[] = DFLT_FRM_MAC;
  unsigned char bcast_mac[] = BCAST_MAC;
  unsigned char mstr_mac[] = DEV_MAC;
  unsigned char mac[6];
  unsigned char slv_mac[6];
  unsigned char pmacs[NMAX_MACS][6];
  int tag_macs[NMAX_MACS];

  char cmnd[5],cnfg_fn[128],c,ans[12],adata[20],adata2[20];
  int i,j,l,n,lp,pg,cnt,lines,exit_lp,dhelp,errno,lpcnt;
  int type,err_type;
  int rd_rtn_pkt;
  int try,nlimit,doff;
  FILE *fp;
  unsigned int x;
  float fvalue,scale;
  int slv_mtch,unt_mtch,rsp_macs,unt,rsp;
  int npol;
  int cur_dev;


  /*
   *
   * Initialization
   */
  fp=fopen("mac_addr_slave.hex","r");
  fscanf(fp,"%02X-%02X-%02X-%02X-%02X-%02X",&slv_mac[0],&slv_mac[1],&slv_mac[2],&slv_mac[3],&slv_mac[4],&slv_mac[5]);
  fclose(fp);

  lpcnt=1;
  max=0;
  if(numset>max)max=numset;
  if(numval>max)max=numval;
  if(numcmd>max)max=numcmd;
  for(i=0;i<JMB_DATA_LEN;i++)wbuf[i]=0;
  eth_open("vme0");
  bcast  = eth_register_mac(bcast_mac);
  uninit = eth_register_mac(dflt_mac);
  slave  = eth_register_mac(slv_mac);
  eth_reset();
  eth_read_timeout(RD_TMO_short);
  /*
   *
   * Scanning for boards
   */
  cur_mac = bcast;
  npol = poll_macs(pmacs);
  slv_mtch=unt_mtch=rsp_macs=0;
  printf("received %d packets\n",npol);
  if(npol>0){
    printf("The following MAC devices were found:\n");
    for(n=0;n<npol;n++){
      printf("                      %d: ",n);
      print_mac(pmacs[n]);
      if(maccmp(pmacs[n],slv_mac)==0){
        slv_mtch++;
        tag_macs[n]=slave;
      }
      else if(maccmp(pmacs[n],dflt_mac)==0){
        unt_mtch++;
        unt=n;
        tag_macs[n]=uninit;
      }
      else {
        tag_macs[n]= -1;
        rsp=n;
        rsp_macs++;
      }
    }
    printf("Registering new devices.\n");
    for(n=0;n<npol;n++){
      if(tag_macs[n]==-1) tag_macs[n]= eth_register_mac(pmacs[n]);
    }
    if(rsp_macs>0){
      cur_mac = tag_macs[rsp];
      cur_dev = rsp;
      printf("Current MAC set to     : ");print_mac(pmacs[rsp]);
      strcpy(cmnd,"help");
    }
    else {
      printf("Current Mac is not initialized.  You must choose one.\n");
      strcpy(cmnd,"sel_dev");
    }
  }
  else {
    printf("No responses found!  Check your setup.\n");
    exit(1);
  }

  /*
   *
   * Initialize VME structures
   * 
   */
  vc.cw.tg.tt  = SNGL;
  vc.cw.tg.dsz = D16;
  vc.cw.tg.wrt = WRITE;
  vc.cw.tg.asz = A24;
  vc.cw.tg.dt = No_Dly;
  vc.cw.tg.at = 0x00;
  //  vc.cw.fl.pg = 0; // alternative  method
  //  vc.cw.fl.sp = 0;
  //  vc.cw.fl.lk = 0;
  //  vc.cw.fl.cr = 0;
  //  vc.cw.fl.ud = 0;
  if(vc.cw.tg.asz<A40){
    vc.addr_l = slot<<slot_shft[vc.cw.tg.asz];
    vc.addr_h = 0x00000000;
  }
  else {
    vc.addr_l = 0x00000000;
    vc.addr_h = slot<<(slot_shft[vc.cw.tg.asz]-32);
  }
  vc.dc = 0;
  dcnt = 0;
  vc.dp = idata;
  vc.delay = 0;
  vc.uam = 0x10;
  dev = 0;
  off = 0;
  aoff = 0;
  tt = vc.cw.tg.tt;
  dsz = vc.cw.tg.dsz;
  wrt = vc.cw.tg.wrt;
  asz = vc.cw.tg.asz;
  dt = vc.cw.tg.dt;
  update_am();
  rd_rtn_pkt = 1;

  /*
   *
   * Command parsing
   * 
   */
  strcpy(cmnd,"h");
  while(strcmp(cmnd,"q")!=0){
    if(strcmp(cmnd,"h")==0){
      dhelp =1;
    }
    else if(strcmp(cmnd,"ron")==0){
      rd_rtn_pkt = 1;
    }
    else if(strcmp(cmnd,"roff")==0){
      rd_rtn_pkt = 0;
    }
    else if(strcmp(cmnd,"clr")==0){
      numwrds = 0;
    }
    else if(strcmp(cmnd,"tt")==0){
      printf("Choose Transfer Type\n");
      for(i=0;i<4;i++){
        printf("%d:%s\n",i,trns_typ[i]);
      }
      printf("Trns Type> ");
      scanf("%d",&tt);
      vc.cw.tg.tt = tt;
      update_am();
    }
    else if(strcmp(cmnd,"SNGL")==0){
      vc.cw.tg.tt = tt = SNGL;
      update_am();
    }
    else if(strcmp(cmnd,"BLOCK")==0){
      vc.cw.tg.tt = tt = BLOCK;
      update_am();
      printf("Number of Transfers?> ");
      scanf("%d",&vc.dc);
    }
    else if(strcmp(cmnd,"RMW")==0){
      vc.cw.tg.tt = tt = RMW;
      update_am();
    }
    else if(strcmp(cmnd,"UNALG")==0){
      vc.cw.tg.tt = tt = UNALG;
      update_am();
    }
    else if(strcmp(cmnd,"dsz")==0){
      printf("Choose Data Size\n");
      for(i=0;i<4;i++){
        printf("%d:%s\n",i,data_sz[i]);
      }
      printf("Data Size> ");
      scanf("%d",&dsz);
      vc.cw.tg.dsz = dsz;
      update_am();
    }
    else if(strcmp(cmnd,"D08")==0){
      vc.cw.tg.dsz = dsz = D08;
      update_am();
    }
    else if(strcmp(cmnd,"D16")==0){
      vc.cw.tg.dsz = dsz = D16;
      update_am();
    }
    else if(strcmp(cmnd,"D32")==0){
      vc.cw.tg.dsz = dsz = D32;
      update_am();
    }
    else if(strcmp(cmnd,"D64")==0){
      vc.cw.tg.dsz = dsz = D64;
      update_am();
    }
    else if(strcmp(cmnd,"wrt")==0 || strcmp(cmnd,"W")==0){
      vc.cw.tg.wrt = wrt = WRITE;
    }
    else if(strcmp(cmnd,"rd")==0 || strcmp(cmnd,"R")==0){
      vc.cw.tg.wrt = wrt = READ;
    }
    else if(strcmp(cmnd,"asz")==0){
      printf("Choose Address Size\n");
      for(i=1;i<6;i++){
        printf("%d:%s\n",i,addr_sz[i]);
      }
      printf("Addr Size> ");
      scanf("%d",&asz);
      vc.cw.tg.asz = asz;
      update_addr();
      update_am();
    }
    else if(strcmp(cmnd,"A16")==0){
      vc.cw.tg.asz = asz = A16;
      update_addr();
      update_am();
    }
    else if(strcmp(cmnd,"A24")==0){
      vc.cw.tg.asz = asz = A24;
      update_addr();
      update_am();
    }
    else if(strcmp(cmnd,"A32")==0){
      vc.cw.tg.asz = asz = A32;
      update_addr();
      update_am();
    }
    else if(strcmp(cmnd,"A40")==0){
      vc.cw.tg.asz = asz = A40;
      update_addr();
      update_am();
    }
    else if(strcmp(cmnd,"A64")==0){
      vc.cw.tg.asz = asz = A64;
      update_addr();
      update_am();
    }
    else if(strcmp(cmnd,"dt")==0){
      printf("Choose Delay Type\n");
      for(i=0;i<7;i++){
        printf("%d:%s\n",i,dly_typ[i]);
      }
      printf("Delay Type> ");
      scanf("%d",&dt);
      vc.cw.tg.dt = dt;
    }
    else if(strcmp(cmnd,"at")==0){
      exit_lp = 0;
      while(!exit_lp){
        printf("Choose Access Type\n");
        printf("Supv/NPrv,Data/Prgm,Lock,CCSR,UDAM,reset,exit\n");
        printf("Access Type (s,n,d,p,l,c,u,r, or e)> ");
        scanf("%s",ans);
        switch(ans[0]){
        case 's':
          vc.cw.fl.sp = 1;
          break;
        case 'n':
          vc.cw.fl.sp = 0;
          break;
        case 'd':
          vc.cw.fl.pg = 0;
          break;
        case 'p':
          vc.cw.fl.pg = 1;
          break;
        case 'l':
          vc.cw.fl.lk = 1;
          break;
        case 'c':
          vc.cw.fl.cr = 1;
          break;
        case 'u':
          vc.cw.fl.ud = 1;
          break;
        case 'r':
          vc.cw.fl.sp = 0;
          vc.cw.fl.pg = 0;
          vc.cw.fl.lk = 0;
          vc.cw.fl.cr = 0;
          vc.cw.fl.ud = 0;
          break;
        case 'e':
          exit_lp = 1;
          break;
        default:
          printf("not a valid option. -- try again\n");
          break;
        }
        display(0);
      }
      update_am();
    }
    else if(strcmp(cmnd,"s")==0){
      printf("Enter Slot Number\n");
      printf("Slot Number> ");
      scanf("%d",&slot);
      if(asz<A40){
        vc.addr_l = vc.addr_l&slot_mask_l[asz] | slot<<slot_shft[asz];
        vc.addr_h = 0x00000000;
      }
      else {
        vc.addr_l = 0x00000000;
        vc.addr_h = vc.addr_h&slot_mask_h[asz] | slot<<(slot_shft[asz]-32);
      }
    }
    else if(strcmp(cmnd,"dev")==0){
      printf("Choose the Slave Device\n");
      for(i=0;i<6;i++){
        printf("%d:%s\n",i,slv_dev[i]);
      }
      printf("Slave Dev> ");
      scanf("%d",&dev);
      vc.addr_l = vc.addr_l&dev_mask[asz] | (b_addr[dev]+aoff)&~dev_mask[asz];
    }
    else if(strcmp(cmnd,"off")==0){
      printf("Enter the Address offset\n");
      printf("Offset > ");
      scanf("%d",&off);
      aoff = off*(inc_sz[dsz]);
      vc.addr_l = vc.addr_l&dev_mask[asz] | (b_addr[dev]+aoff)&~dev_mask[asz];
    }
    else if(strcmp(cmnd,"inc")==0 || strcmp(cmnd,"I")==0){
      off++;
      aoff+=inc_sz[dsz];
      vc.addr_l = vc.addr_l&dev_mask[asz] | (b_addr[dev]+aoff)&~dev_mask[asz];
    }
    else if(strcmp(cmnd,"dec")==0 || strcmp(cmnd,"D")==0){
      off--;
      aoff-=inc_sz[dsz];
      vc.addr_l = vc.addr_l&dev_mask[asz] | (b_addr[dev]+aoff)&~dev_mask[asz];
    }
    else if(strcmp(cmnd,"data")==0){
      dcnt = 0;
      printf("Enter Data end with \"eod\"\n");
      printf("DATA (1-16Hex Digits |dc=%03d> ",dcnt);
      scanf("%s",adata);
      while(strcmp(adata,"eod")!=0){
        if(non_hex_digits(adata)){
          printf("\n\aInvalid hex number. Try again\n");
          sleep(2);
        }
        else {
          if(dsz!=D64){
            sscanf(adata,"%8x",&idata[dcnt++]);
          }
          else {
            l=strlen(adata)-1;
            for(i=15;i>=0;i--){
              adata2[i] = (l<0)?'0':adata[l--];
            }
            adata2[16]='\0';
            sscanf(adata2,"%8x%8x",&idata[2*dcnt],&idata[2*dcnt+1]);
            dcnt++;
          }
        }
        display(0);
        printf("Enter Data end with \"eod\"\n");
        printf("DATA (1-16Hex Digits |dc=%03d> ",dcnt);
        scanf("%s",adata);
      }
    }
    else if(strcmp(cmnd,"dly")==0){
        printf("Enter Delay Value with units:\n");
        printf("Default is miliseconds:\n");
        printf("Delay (#.# us,ms,s,hr> ");
        scanf("%f %s",&fvalue,ans);
        if(strcmp(ans,"us")==0){
          scale = 1E-6;
        }
        else if(strcmp(ans,"ms")==0){
          scale = 1E-3;
        }
        else if(strcmp(ans,"s")==0){
          scale = 1;
        }
        else if(strcmp(ans,"hr")==0){
          scale = 3600;
        }
        else {
          scale = 1E-3;
        }
        if(dt!=0){
          vc.delay = (unsigned int)(fvalue*scale/clk_units[dt]);
        }
        else {
          vc.delay = (unsigned int)(fvalue*scale/clk_units[6]);
	}
    }
    else if(strcmp(cmnd,"uam")==0){
      printf("Enter User Defined Address Modifier\n");
      printf("UD AM (hex 10-1F)> ");
      scanf("%2X",&vc.uam);
    }
    else if(strcmp(cmnd,"lp")==0){
      printf("Enter Loop Count> ");
      scanf("%d",&lpcnt);
    }
    else if(strcmp(cmnd,"ex")==0){
      wbuf[0]=0x20;
      wbuf[1]=VME_Cmds;
      wbuf[2]=0x00;
      wbuf[3]=0x01;
      wbuf[4]=vc.cw.full>>8;
      wbuf[5]=vc.cw.full&0xFF;
      nwbuf=6;
      if(dt >0){
        if(dt<4){
          wbuf[nwbuf++] = (vc.delay>>8)&0xFF;
          wbuf[nwbuf++] = vc.delay&0xFF;
	}
        else {
          wbuf[nwbuf++] = (vc.delay>>24)&0xFF;
          wbuf[nwbuf++] = (vc.delay>>16)&0xFF;
          wbuf[nwbuf++] = (vc.delay>>8)&0xFF;
          wbuf[nwbuf++] = vc.delay&0xFF;
        }
      }
      else {
        if(vc.cw.fl.ud){
          wbuf[nwbuf++] = 0x00;
          wbuf[nwbuf++] = vc.uam&0xFF;
        }
        switch(asz){
	case A16:
	  wbuf[nwbuf++] = (vc.addr_l>>8)&0xFF;
	  wbuf[nwbuf++] = vc.addr_l&0xFF;
	  break;
	case A24:
	  wbuf[nwbuf++] = 0x00;
	  wbuf[nwbuf++] = (vc.addr_l>>16)&0xFF;
	  wbuf[nwbuf++] = (vc.addr_l>>8)&0xFF;
	  wbuf[nwbuf++] = vc.addr_l&0xFF;
	  break;
	case A32:
	  wbuf[nwbuf++] = (vc.addr_l>>24)&0xFF;
	  wbuf[nwbuf++] = (vc.addr_l>>16)&0xFF;
	  wbuf[nwbuf++] = (vc.addr_l>>8)&0xFF;
	  wbuf[nwbuf++] = vc.addr_l&0xFF;
	  break;
	case A40:
	  wbuf[nwbuf++] = 0x00;
	  wbuf[nwbuf++] = vc.addr_h&0xFF;
	  wbuf[nwbuf++] = (vc.addr_l>>24)&0xFF;
	  wbuf[nwbuf++] = (vc.addr_l>>16)&0xFF;
	  wbuf[nwbuf++] = (vc.addr_l>>8)&0xFF;
	  wbuf[nwbuf++] = vc.addr_l&0xFF;
	  break;
	case A64:
 	  wbuf[nwbuf++] = (vc.addr_h>>24)&0xFF;
	  wbuf[nwbuf++] = (vc.addr_h>>16)&0xFF;
	  wbuf[nwbuf++] = (vc.addr_h>>8)&0xFF;
	  wbuf[nwbuf++] = vc.addr_h&0xFF;
	  wbuf[nwbuf++] = (vc.addr_l>>24)&0xFF;
	  wbuf[nwbuf++] = (vc.addr_l>>16)&0xFF;
	  wbuf[nwbuf++] = (vc.addr_l>>8)&0xFF;
	  wbuf[nwbuf++] = vc.addr_l&0xFF;
	  break;
	default:
          printf("Error in address size.\n");
          printf("Command is not being executed\n");
          errno = 1;
          sleep(2);
	  break;
        }
        if(tt == BLOCK){
          wbuf[nwbuf++] = (vc.dc>>8)&0xFF;
          wbuf[nwbuf++] = vc.dc&0xFF;
          nlimit = (dsz==D64)?2*vc.dc:vc.dc;
          doff = off;
        }
        else {
          nlimit = 1;
          doff = 0;
        }
        if(wrt){
          if(dcnt==0){
            printf("No DATA specified.\n");
            printf("Command is not being executed\n");
            errno = 3;
            sleep(2);
          }
          else {
            for(i=0;i<nlimit;(dsz==D64)?i+=2:i++){
              switch(dsz){
	      case D08:
	        wbuf[nwbuf++] = 0x00;
	        wbuf[nwbuf++] = vc.dp[doff+i]&0xFF;
	        break;
              case D16:
	        wbuf[nwbuf++] = (vc.dp[doff+i]>>8)&0xFF;
	        wbuf[nwbuf++] = vc.dp[doff+i]&0xFF;
	        break;
	      case D32:
	        wbuf[nwbuf++] = (vc.dp[doff+i]>>24)&0xFF;
	        wbuf[nwbuf++] = (vc.dp[doff+i]>>16)&0xFF;
	        wbuf[nwbuf++] = (vc.dp[doff+i]>>8)&0xFF;
	        wbuf[nwbuf++] = vc.dp[doff+i]&0xFF;
	        break;
	      case D64:
	        wbuf[nwbuf++] = (vc.dp[doff+i]>>24)&0xFF;
	        wbuf[nwbuf++] = (vc.dp[doff+i]>>16)&0xFF;
	        wbuf[nwbuf++] = (vc.dp[doff+i]>>8)&0xFF;
	        wbuf[nwbuf++] = vc.dp[doff+i]&0xFF;
	        wbuf[nwbuf++] = (vc.dp[doff+i+1]>>24)&0xFF;
	        wbuf[nwbuf++] = (vc.dp[doff+i+1]>>16)&0xFF;
	        wbuf[nwbuf++] = (vc.dp[doff+i+1]>>8)&0xFF;
	        wbuf[nwbuf++] = vc.dp[doff+i+1]&0xFF;
	        break;
	      default:
                printf("Error in data size.\n");
                printf("Command is not being executed\n");
                errno = 2;
                sleep(2);
	        break;
              }
            }
          }
        }
      }
      if(errno==0){
        for(lp=0;lp<lpcnt;lp++){
	  verrfl=0;
	  no_rsp = 0;
	  numwrds = 0;
	  n=eth_write();
	  if(rd_rtn_pkt){
	    n=0;
	    try = 0;
	    while(n<=6 && try++<1){
	      //            usleep(1000);
	      n=eth_read();
	    }
	    if(n<=6){
	      if(!wrt)no_rsp = 1;
	    }
	    else {
	      type = rbuf[15]&0xff;
	      numwrds = ((rbuf[20]&0x1f)<<8)|(rbuf[21]&0xff);
	      for(i=0;i<2*numwrds;i++)data_wrds[i]=rbuf[22+i];
	      if(type==0xFF){
		verrfl = 1;
		isrc = (rbuf[22]&0xF0)>>4;
		if(isrc==1){
		  verr_type = rbuf[23]&0xFF;
		}
		else if(isrc==2){
		  verr_type =4+((rbuf[23]&0x60)>>5);
		}
		else if(isrc==3){
		  verr_type = 8+rbuf[23]&0xFF;
		}
		else if(isrc==4){
		  verr_type = 12+rbuf[23]&0xFF;
		}
		else {
		  verr_type =16;
		}
		for(i=0;i<2*numwrds;i++)verr_wrds[i]=rbuf[22+i];
		numerr_wrds=numwrds;
		numwrds=0;
		if(isrc==2 && !wrt){
		  n=0;
		  try = 0;
		  while(n<=6 && try++<1){
		    //                  usleep(100000);
		    n=eth_read();
		  }
		  if(n<=6) {
		    no_rsp = 1;
		  }
		  else {
		    numwrds = ((rbuf[20]&0x1f)<<8)|(rbuf[21]&0xff);
		    for(i=0;i<2*numwrds;i++)data_wrds[i]=rbuf[22+i];
		  }
		}
	      }
	      else { // check for error packet after data
		n=0;
		try = 0;
		while(n<=6 && try++<1){
		  //            usleep(1000);
		  n=eth_read();
		}
		if(n>6){
		  numerr_wrds = ((rbuf[20]&0x1f)<<8)|(rbuf[21]&0xff);
		  for(i=0;i<2*numerr_wrds;i++)verr_wrds[i]=rbuf[22+i];
		  verrfl = 1;
		  isrc = (rbuf[22]&0xF0)>>4;
		  if(isrc==1){
		    verr_type = rbuf[23]&0xFF;
		  }
		  else if(isrc==2){
		    verr_type =4+((rbuf[23]&0x60)>>5);
		  }
		  else if(isrc==3){
		    verr_type = 8+rbuf[23]&0xFF;
		  }
		  else if(isrc==4){
		    verr_type = 12+rbuf[23]&0xFF;
		  }
		  else {
		    verr_type =16;
		  }
		}
	      }
	    }
	  }
          usleep(10000);
	}
      }
    }
    else {
      printf("Unknown command, try again...\n");
      sleep(2);
    }
    display(dhelp);
    dhelp =0;
    printf("Cmnd> ");
    scanf("%s",cmnd);
  }
  eth_close();
}


void display(int help)
{
  char cls[] = "\x1B[2J";
  char rtns[] = "\n\n\n\n\n\n\n\n\n\n\n\n";
  char up20[] = "\x1B[20A";
  //clear screen first
  printf("%s%s%s%s",cls,rtns,rtns,up20);  
  disp_v();
  if(help) disp_h();
  return;
}

void disp_v()
{
  char tmpfmt[24];
  int i,j,k;
  printf("---------------------------------------------------------\n");
  if(vc.cw.tg.tt==BLOCK){
    printf("Slot: %2d  %-3s:%-3s:%-5s:%-5s(%4d) %-8s:%u 0x%08X\n",slot,
        addr_sz[vc.cw.tg.asz],data_sz[vc.cw.tg.dsz],trns_typ[vc.cw.tg.tt],
        wrt_rd[vc.cw.tg.wrt],vc.dc,dly_typ[vc.cw.tg.dt],vc.delay,vc.delay);
  }
  else {
    printf("Slot: %2d  %-3s:%-3s:%-5s:%-5s       %-8s:%u 0x%08X\n",slot,
        addr_sz[vc.cw.tg.asz],data_sz[vc.cw.tg.dsz],trns_typ[vc.cw.tg.tt],
        wrt_rd[vc.cw.tg.wrt],dly_typ[vc.cw.tg.dt],vc.delay,vc.delay);
  }
  printf("Access:   %4s+%4s+%4s+%4s+%4s\n",acc_typ_sp[vc.cw.fl.sp],
         acc_typ_pg[vc.cw.fl.pg],acc_typ_lk[vc.cw.fl.lk],
         acc_typ_cr[vc.cw.fl.cr],acc_typ_ud[vc.cw.fl.ud]);
  printf("Addr:     %08X%08X ",vc.addr_h,vc.addr_l);
  if(vc.cw.fl.ud){
    printf("User AM:%02X  ",vc.uam);
  }
  else {
    printf("     AM:%02X  ",cur_am);
  }
  printf("Device:%s\n",slv_dev[dev]);
  printf("Data: %4d:",dcnt);
  if(dcnt==0) printf("no data specified");
  if(dsz==D64){
    for(i=0;(i<dcnt)&&(i<dlmt[dsz]);i++){
      strcpy(tmpfmt,dfmt[dsz]);
      strcat(tmpfmt,dfmt[dsz]);
      printf(tmpfmt,vc.dp[2*i],vc.dp[2*i+1]);
      if((i<dlmt[dsz]-1)&&(i<dcnt-1)){
        printf(",");
      }
    }
  }
  else {
    for(i=0;(i<dcnt)&&(i<dlmt[dsz]);i++){
      printf(dfmt[dsz],vc.dp[i]);
      if((i<dlmt[dsz]-1)&&(i<dcnt-1)){
        printf(",");
      }
    }
  }
  if(dcnt>dlmt[dsz]){
    printf("...\n");
  }
  else {
    printf("\n");
  }
  printf("Rdbk:%5d:",numwrds);
  if(no_rsp==1){
    printf("no response");
  }
  else if(numwrds==0) {
    printf("no data read");
  }
  if(dsz==D64){
    for(i=0;(i<numwrds/4)&&(i<dlmt[dsz]);i++){
      for(j=0;j<8;j++)printf("%02X",data_wrds[8*i+j]);
      if((i<dlmt[dsz]-1)&&(i<(numwrds/4-1))){
        printf(",");
      }
    }
  }
  else if(dsz==D32){
    for(i=0;(i<numwrds/2)&&(i<dlmt[dsz]);i++){
      for(j=0;j<4;j++)printf("%02X",data_wrds[4*i+j]);
      if((i<dlmt[dsz]-1)&&(i<(numwrds/2-1))){
        printf(",");
      }
    }
  }
  else if(dsz==D16){
    for(i=0;(i<numwrds)&&(i<dlmt[dsz]);i++){
      for(j=0;j<2;j++)printf("%02X",data_wrds[2*i+j]);
      if((i<dlmt[dsz]-1)&&(i<numwrds-1)){
        printf(",");
      }
    }
  }
  else {
    for(i=0;(i<2*numwrds)&&(i<dlmt[dsz]);i++){
      printf("%02X",data_wrds[i]);
      if((i<dlmt[dsz]-1)&&(i<(2*numwrds-1))){
        printf(",");
      }
    }
  }
  if(i>dlmt[dsz]){
    printf("...\n");
  }
  else {
    printf("\n");
  }
  if(verrfl){
    printf("%10s:",verr_types[verr_type]);
    for(i=0;(i<numerr_wrds)&&(i<10);i++){
      for(j=0;j<2;j++)printf("%02X",verr_wrds[2*i+j]);
      if((i<9)&&(i<numerr_wrds-1)){
        printf(",");
      }
    }
    printf("\n");
  }
  else {
    printf("  no Error:\n");
  }
  printf("---------------------------------------------------------\n");
  return;
}


void disp_h()
{
char *ptr1a,*ptr2a,*ptr3a;
char *ptr1b,*ptr2b,*ptr3b;
int l;
  printf("      Settings             Values                Commands\n");
  printf("------------------------------------------------------------------\n");
  for(l=0;l<max;l++){
    ptr1a = (l<numset)?settings[l]:blank;
    ptr2a = (l<numval)?values[l]:blank;
    ptr3a = (l<numcmd)?cmds[l]:blank;
    ptr1b = (l<numset)?s_desc[l]:blank;
    ptr2b = (l<numval)?v_desc[l]:blank;
    ptr3b = (l<numcmd)?c_desc[l]:blank;
    printf("%-7s%-15s%-7s%-15s%-7s%-15s\n",
           ptr1a,ptr1b,ptr2a,ptr2b,ptr3a,ptr3b);
  }
}

int non_hex_digits(char *s)
{
  int i,n;
  n=i=0;
  while(s[i]!='\0'){
    if(!isxdigit(s[i++]))n++;
  }
  return(n);
}

void update_addr()
{
      if(vc.cw.tg.asz<A40){
        vc.addr_l = slot<<slot_shft[vc.cw.tg.asz];
        vc.addr_h = 0x00000000;
      }
      else {
        vc.addr_l = 0x00000000;
        vc.addr_h = slot<<(slot_shft[vc.cw.tg.asz]-32);
      }
      vc.addr_l = vc.addr_l&dev_mask[vc.cw.tg.asz] | (b_addr[dev]+aoff)&~dev_mask[vc.cw.tg.asz];
}



void update_am()
{
    if(vc.cw.fl.cr){
      cur_am = 0x2F;
    }
    else if(vc.cw.tg.asz == A16){
      if(vc.cw.fl.lk){
        cur_am = 0x2C;
      }
      else if(vc.cw.fl.sp){
        cur_am = 0x2D;
      }
      else {
        cur_am = 0x29;
      }
    }
    else if(vc.cw.tg.asz == A24){
      if(vc.cw.fl.lk){
        cur_am = 0x32;
      }
      else if(vc.cw.fl.sp){
        if(vc.cw.tg.tt == BLOCK){
          if(vc.cw.tg.dsz == D64){
            cur_am = 0x3C;
          }
	  else {
            cur_am = 0x3F;
          }
        }
        else{
          if(vc.cw.fl.pg){
            cur_am = 0x3E;
          }
	  else {
            cur_am = 0x3D;
          }
        }
      }
      else {
        if(vc.cw.tg.tt == BLOCK){
          if(vc.cw.tg.dsz == D64){
            cur_am = 0x38;
          }
	  else {
            cur_am = 0x3B;
          }
        }
        else{
          if(vc.cw.fl.pg){
            cur_am = 0x3A;
          }
	  else {
            cur_am = 0x39;
          }
        }
      }
    }
    else if(vc.cw.tg.asz == A32){
      if(vc.cw.fl.lk){
        cur_am = 0x05;
      }
      else if(vc.cw.fl.sp){
        if(vc.cw.tg.tt == BLOCK){
          if(vc.cw.tg.dsz == D64){
            cur_am = 0x0C;
          }
	  else {
            cur_am = 0x0F;
          }
        }
        else{
          if(vc.cw.fl.pg){
            cur_am = 0x0E;
          }
	  else {
            cur_am = 0x0D;
          }
        }
      }
      else {
        if(vc.cw.tg.tt == BLOCK){
          if(vc.cw.tg.dsz == D64){
            cur_am = 0x08;
          }
	  else {
            cur_am = 0x0B;
          }
        }
        else{
          if(vc.cw.fl.pg){
            cur_am = 0x0A;
          }
	  else {
            cur_am = 0x09;
          }
        }
      }
    }
    else if(vc.cw.tg.asz == A40){
      if(vc.cw.fl.lk){
        cur_am = 0x35;
      }
      else if(vc.cw.tg.tt == BLOCK){
        cur_am = 0x37;
      }
      else {
        cur_am = 0x34;
      }
    }
    else if(vc.cw.tg.asz == A64){
      if(vc.cw.fl.lk){
        cur_am = 0x04;
      }
      else if(vc.cw.tg.tt == BLOCK){
        if(vc.cw.tg.dsz == D64){
          cur_am = 0x00;
        }
        else {
          cur_am = 0x03;
        }
      }
      else {
        cur_am = 0x01;
      }
    }
}
